"""JSON module.

Problems with classical json that are solved:
- When loading, int and float values are strings. So `read` function uses `NumbersDecoder` by default.
- When dumping, object that can be converted to string or list are not converted.
 So `write` function converts objects to string and iterable to list.


Examples:
---------
Write to file:
```
from dh5 import jsn
data = {'a': 1}
jsn.write(path, data)
```

Open file:

```
from dh5 import jsn
data = jsn.read(path)
```
"""
from .methods import read, write  # noqa: F401



# flake8: noqa: F401 # pylint: disable=E0401
import typing as _t
from copy import deepcopy as _deepcopy

from . import dh5_types
from .__config__ import __version__
from .dh5_class import DH5

if _t.TYPE_CHECKING:
    from pathlib import Path  # pragma: no cover


def load(
    filepath_or_data: _t.Optional[_t.Union[str, dict, "Path"]] = None,
    /,
    mode: _t.Optional[_t.Literal["r", "w", "a", "w=", "a="]] = None,
    *,
    filepath: _t.Optional[_t.Union[str, "Path"]] = None,
    save_on_edit: bool = False,
    overwrite: _t.Optional[bool] = None,
    data: _t.Optional[dict] = None,
    open_on_init: _t.Optional[bool] = None,
    **kwds,
):
    """Open H5 file in read/write mode.


    Args:
        filepath_or_data (str|dict, optional): either filepath, either data as dict.
        filepath (str|Path, optional): filepath to load. Defaults to None.
        save_on_edit (bool, optional): Save data as soon as you changed it.
            Defaults to False. And data should be saved using `save()` method.
        overwrite: If file exists, overwrite it. Defaults to raise error is file exist.
        data (Optional[dict], optional):
            Data to load. If data provided, file . Defaults to None.
        open_on_init (Optional[bool], optional): open_on_init. Defaults to True.

    """

    return DH5(
        filepath_or_data,
        mode=mode,
        filepath=filepath,
        save_on_edit=save_on_edit,
        data=data,
        open_on_init=open_on_init,
        overwrite=overwrite,
        **kwds,
    )


def save(
    data: dict,
    filepath: str,
    mode: _t.Literal["w", "a"] = "w",
    overwrite: _t.Optional[bool] = None,
    **kwds,
):
    """Save data to H5 file.

    Args:
        filepath (str): The path to the H5 file.
        data (dict): The data to be saved.
        **kwds: Additional keyword arguments to be passed to the DH5 constructor.

    Returns:
        DH5: DH5 object with data.
    """

    dh5 = DH5(data=_deepcopy(data), **kwds, filepath=filepath, overwrite=overwrite, mode=mode)
    return dh5.save(filepath=filepath)



"""Data transformation utils."""

import json
from typing import Iterable, Optional, Sized

import numpy as np

from ..types import DICT_OR_LIST_LIKE


def np_array_check(lst, size: Optional[int] = None) -> int:
    """Check if the list can be converted to np.ndarray.

    Args:
        lst (list|np.ndarray): Any list
        size (int, optional): Expected size. Defaults to None.

    Returns:
        int: size of the list if it can be converted to np.ndarray, -1 otherwise.

    Examples:
    --------
    [[1, 2], [2, 3], [3, 4]] -> Can be converted to np.ndarray. Return 3.
    [1, [2, 3], 4] -> Cannot be converted to np.ndarray. Return -1.

    """
    if isinstance(lst, Iterable) and isinstance(lst, Sized) and not isinstance(lst, str):
        # if isinstance(lst, (list, np.ndarray)):
        if size is not None and size != len(lst):
            return -1  # pragma: no cover
        for lst_elm in lst:
            check = np_array_check(lst_elm, size)
            if check >= 0:
                size = check
            else:
                return -1
        return len(lst)
    if isinstance(lst, (np.number)):
        return 0
    return -1


def transform_to_possible_formats(data: DICT_OR_LIST_LIKE) -> DICT_OR_LIST_LIKE:
    """Transform data to possible formats on setitem.

    If data can be converted to dict or np.ndarray, it will be converted.
    Unless it has attribute __should_not_be_converted__ set to True. Then nothing happens and data
    will be converted during saving.
    """
    if hasattr(data, "__should_not_be_converted__") and data.__should_not_be_converted__ is True:  # type: ignore
        return data

    if hasattr(data, "asdict"):
        data = data.asdict()  # type: ignore

    if hasattr(data, "_asdict"):
        data = data._asdict()  # type: ignore

    if hasattr(data, "asarray"):
        data = data.asarray()  # type: ignore

    if isinstance(data, dict):
        for key, value in data.items():
            data[key] = transform_to_possible_formats(value)
        return data

    if isinstance(data, (tuple, set)):
        data = list(data)

    if isinstance(data, list) and (np_array_check(data) < 0):
        return data

    return data


def transform_on_open(value):
    """Transform data during opening of h5 file."""
    if isinstance(value, bytes):
        value = value.decode()
    if isinstance(value, str) and value.startswith("__json__"):
        return json.loads(value[8:])
    if isinstance(value, str) and value.startswith("__function__"):
        from .transformation_types import function_save

        return function_save.Function(value[12:])
    return value


def transform_not_dict_on_save(value, level=0):
    """Transform data during saving of h5 file if data is not a dict."""
    if isinstance(value, np.ndarray):
        if level == 0:
            return value
        return value.tolist()

    if isinstance(value, (list)) and (np_array_check(value) < 0):
        try:
            if level == 0:
                return "__json__" + json.dumps(value)
            return value
        except TypeError:
            value_transformed = [transform_not_dict_on_save(v, level=level + 1) for v in value]
            return "__json__" + json.dumps(value_transformed)

    if callable(value):
        from .transformation_types import function_save

        return "__function__" + function_save.function_to_str(value)

    return value



"""DH5 module. Main class is DH5. It's a dictionary that is synchronized with .h5 file.

Also this module contains h5py_utils that allows to save and load dict from a h5 file.

Examples:
---------
>>> from dh5 import DH5
>>> data = DH5("some_file.h5", mode='w')
>>> data["some_key"] = "some_value"
>>> data.save()

>>> data = DH5("some_file.h5", mode='r')
>>> data["some_key"] # -> "some_value"
>>> data["some_key"] = "some_other_value" # -> KeyError: Cannot change value in read mode

"""


# flake8: noqa: F401
from .main import DH5



"""Contains all custom JSON encoders used."""
import json


class StringEncoder(json.JSONEncoder):
    """Classical JSONEncoder will not try to convert objects to a str. This does."""

    def default(self, o):
        """Additionally encode all iterators as a list."""
        if hasattr(o, "__iter__"):
            return [self.default(obj) for obj in iter(o)]
        return str(o)



"""All types used for typing inside dh5 module."""

from typing import Callable, Optional, Protocol, Union

import numpy as np


class ClassWithAsdict(Protocol):
    """Any class with predefined `_asdict` attribute.
    `_asdict` class should return a dictionary with only list and dict.
    It should not be a dict of other classes
    """

    def asdict(self) -> dict:  # type: ignore
        """Return internal data as a dictionary."""


class ClassWithAsarray(Protocol):
    """
    Any class with predefined `asarray` attribute.
    `asarray` class should return a np.ndarray.
    """

    def asarray(self) -> Union[np.ndarray, list]:  # type: ignore
        """Return internal data as an array."""


DICT_OR_LIST_LIKE = Optional[
    Union[
        dict,
        list,
        np.ndarray,
        ClassWithAsdict,
        ClassWithAsarray,
        np.int_,
        np.float_,
        float,
        int,
        str,
        Callable,
    ]
]
RIGHT_DATA_TYPE = Union[dict, np.ndarray, np.int_, np.float_, float, int]



# flake8: noqa: D100
import json
from typing import Optional, Type, Union
from .decoders import NumbersDecoder
from .encoders import StringEncoder
from ..path import Path, get_file_path


def write(
    file: Union[str, "Path"],
    data: dict,
    path: Optional[str] = None,
    encoder: Type[json.JSONEncoder] = StringEncoder,
) -> None:
    """Write data to a file. Sort keys.

    Args:
        file (str): file name or path.
        data (dict): data.
        path (Optional[str], optional): Path to the folder with file. Defaults to None.

    Examples:
        >>> write(path, data)
    """
    file = get_file_path(file, path=path, extension=".json")
    file.dirname.makedirs()
    with open(file, "w", encoding="utf-8") as outfile:
        json.dump(data, outfile, sort_keys=True, indent=4, cls=encoder)


def read(
    file: Union[str, "Path"],
    path: Optional[str] = None,
    decoder: Type[json.JSONDecoder] = NumbersDecoder,
) -> dict:
    """Load a data from file at path. Decode number by default.

    Args:
        file (Union[str, Path]): file name or path to file.
        path (Optional[str], optional): path to folder with file. Defaults to None.

    Returns:
        dict: Data loaded from file.

    Examples:
        >>> read(path)

    """
    file = get_file_path(file, path=path, extension=".json")

    with open(file, encoding="utf-8") as json_file:
        return json.load(json_file, cls=decoder)



# flake8: noqa: D100, D101, D102
import os
from typing import Optional

import h5py
import numpy as np


class SyncNp(np.ndarray):
    __filename__: Optional[str] = None
    __filekey__: Optional[str] = None
    __should_not_be_converted__ = True
    __save_on_edit__: bool = False
    __last_changes__: Optional[list] = None
    __should_initialized__: bool = False

    def __new__(cls, data):
        if isinstance(data, list):
            data = np.array(data)
        if isinstance(data, tuple):
            data = np.zeros(shape=data)
        if isinstance(data, SyncNp):
            return data
        data = data.view(cls)
        return data

    def __init__filepath__(self, *, filepath: str, filekey: str, save_on_edit: bool = False, **_):
        self.__filename__ = filepath
        self.__filekey__ = filekey
        self.__save_on_edit__ = save_on_edit
        if self.__last_changes__ is None:
            self.__last_changes__ = []
        self.__should_initialized__ = True
        if self.__save_on_edit__:  # or not self.__initialized__:
            self.save(only_update=False)

    def __setitem__(self, __key, __value):  # type: ignore
        if self.__last_changes__ is None:
            self.__last_changes__ = []
        self.__last_changes__.append(__key)

        super().__setitem__(__key, __value)

        if self.__save_on_edit__:
            self.save(only_update=True)

    def save(self, only_update=True):
        if self.__last_changes__ is None:
            return self

        if not self.__filename__ or not self.__filekey__:
            raise ValueError("Cannot save changes without filename and filekey provided")

        if os.path.dirname(self.__filename__) and not os.path.exists(
            os.path.dirname(self.__filename__)
        ):
            os.makedirs(os.path.dirname(self.__filename__), exist_ok=True)

        if not only_update:
            self.__should_initialized__ = False
            with h5py.File(self.__filename__, "a") as file:
                if self.__filekey__ in file:
                    del file[self.__filekey__]
                file[self.__filekey__] = np.asarray(self)

        with h5py.File(self.__filename__, "a") as file:
            if self.__filekey__ not in file or self.__should_initialized__:
                if self.__filekey__ in file:
                    del file[self.__filekey__]
                file[self.__filekey__] = np.asarray(self)
                self.__should_initialized__ = False

            for key in self.__last_changes__:
                file[self.__filekey__][key] = self[key]  # type: ignore

        self.__last_changes__ = []
        return self

    def asarray(self):
        return self.view(np.ndarray)



# flake8: noqa: D100
import inspect
import logging
from typing import Any


class Function:
    """Function class contains code of a function that can be evaluated.

    Examples:
    --------
    >>> func = Function("def func(a, b): return a + b")
    >>> func.code # -> "def current_func(a, b): return a + b"
    >>> func.eval(1, 2) # -> 3

    Function is evaluated on first eval call and not before.

    Moreover, __call__ method can be used if the function was already evaluated:

    >>> func = Function("def func(a, b): return a + b")
    >>> func.code # -> "def current_func(a, b): return a + b"
    >>> func(1, 2) # -> TypeError("Function was never evaluated.")
    >>> func.eval(1, 2) # -> 3
    >>> func(1, 2) # -> 3

    """

    code: str = ""
    func = None
    original_name: str = ""

    _read_only = True

    def __init__(self, code: str):
        """Create a `Function` from the code provided.

        Args:
            code (str): Code of the function.
        """
        index_name = code.find("(")
        if index_name == -1:
            return
        self.original_name = code[4:index_name]
        self.code = "def current_func" + code[index_name:]
        if not self.code.endswith("\n"):
            self.code += "\n"

    def _evaluate_func(self):
        """Evaluate the function code and store it in the self.func."""
        if self.code == "":
            raise ValueError(
                f"Function {self.original_name} cannot be loaded because the code is empty."
            )

        try:
            cc = compile(self.code, "<string>", "single")  # noqa: DUO110
            eval(cc)  # pylint: disable=W0123 # noqa: DUO104
        except SyntaxError:
            logging.warning(
                "Function %s cannot be loaded because unexpected SyntaxError.", self.original_name
            )
            return

        self.func = locals().get("current_func")

    def eval(self, *args, **kwds):
        """Evaluate the function and return the result."""
        if self.func is None:
            self._evaluate_func()
        if self.func is None:
            raise SyntaxError(f"Cannot call run a function defined by:\n{self.code}")
        return self.func(*args, **kwds)  # type: ignore

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        """Evaluate the function and return the result if the function was already compiled."""
        if self.func is None:
            raise ImportError(
                "Function was never evaluated. On the first run use the eval(...) method instead."
            )
        return self.func(*args, **kwds)  # type: ignore

    def __str__(self):
        return f"Function: {self.code}"  # pragma: no cover


def function_to_str(func):
    """Return the code source of the function using inspect library."""
    if inspect.isfunction(func):
        return inspect.getsource(func)

    name = func.__name__ if hasattr(func, "__name__") else ""
    raise ValueError(f"Function {name} must be a function and not a module or a class.")



"""Contains all custom JSON decoders used."""

import json


class NumbersDecoder(json.JSONDecoder):
    """Decode float and int."""

    def decode(self, s):  # pylint: disable=W0221 # type: ignore
        """Decode using the default decoder and then decode numbers."""
        result = super().decode(s)
        return self._decode(result)

    def _decode(self, obj):
        """Decode numbers if obj is one, or look inside if obj is a list or a dict."""
        if isinstance(obj, str):
            try:
                obj_check = obj[1:] if obj.startswith("-") else obj
                if obj_check.isnumeric():
                    return int(obj)
                if obj_check.count(".") == 1 and obj_check.replace(".", "").isnumeric():
                    return float(obj)
                return obj
            except ValueError:  # pragma: no cover
                return obj  # pragma: no cover
        if isinstance(obj, dict):
            return {k: self._decode(v) for k, v in obj.items()}
        if isinstance(obj, list):
            return [self._decode(v) for v in obj]
        return obj



"""Path module.

Functionality added to classical Path form pathlib:

- sum 2 path or path and str. Action: sum 2 str
- makedirs. Action: os.makedirs
- make_extension
- dirname
- basename


Examples:
---------
>>> from dh5.path import Path
>>> path = Path("some_folder")
>>> path.makedirs()
>>> path = path / "some_file" # -> "some_folder/some_file"
>>> path = path + ".json" # -> "some_folder/some_file.json"
>>> path.basename # -> "some_file.json"

"""

from .path_class import Path, get_file_path  # noqa: F401



"""DH5 class is a dictionary that is synchronized with .h5 file."""

import logging
import os
from copy import deepcopy
from functools import wraps
from pathlib import Path
from typing import Any, Dict, Iterable, Literal, Optional, Set, TypeVar, Union, overload

from ..errors import ReadOnlyKeyError
from ..types import DICT_OR_LIST_LIKE
from . import h5py_utils
from .data_transformation import transform_to_possible_formats
from .dict_structure import get_keys_structure, output_dict_structure
from .internal_classes import NotLoaded

# from ..utils import


def editing(func):
    """If a function changes the data it should be saved.
    It's a wrapper for such function.
    """

    @wraps(func)
    def run_func_and_clean_precalculated_results(self, *args, **kwargs):
        self._last_data_saved = False  # pylint: disable=W0212
        res = func(self, *args, **kwargs)
        self._clean_precalculated_results()  # pylint: disable=W0212
        if self._save_on_edit:  # pylint: disable=W0212
            self.save(only_update=True)
        return res

    return run_func_and_clean_precalculated_results


_T = TypeVar("_T")
_SELF = TypeVar("_SELF", bound="DH5")


class DH5:
    """DH5 - Dict that is synchronized with .h5 file.

    # Usage and initialization
    DH5 can open files in 3 different modes:

    - 'r' - Read mode. No data chan be modified.
    - 'w' - Write mode. If file exists it will be overwritten. And you have full control on data.
    - 'a' - Append mode. If file exists it will be opened. And you have full control on data.

    To overwrite file use `open_overwrite` method or `mode="w"` with `overwrite=True`.


    # Examples
        >>> sd = DH5('somedata.h5', 'w')
        >>> sd['a'] = 5
        >>> sd.save()

        >>> sd_read = DH5('somedata.h5', 'r')
        >>> sd_read['a']
        5
        >>> sd_read.a
        5

        >>> sd_append = DH5('somedata.h5', 'a')
        >>> sd_append['b'] = 6
        >>> sd_append.save()

        >>> sd_read = DH5('somedata.h5', 'r')
        >>> sd_read['a'], sd_read['b']
        (5, 6)

    """

    _repr: Optional[str] = None
    _default_attr = ["get", "items", "keys", "pop", "update", "values", "save"]
    _last_data_saved: bool = False
    _filepath: Optional[str] = None
    _read_only: Union[bool, Set[str]]
    _raise_file_locked_error: bool = False
    _retry_on_file_locked_error: int = 5
    _last_time_data_checked: float = 0
    _file_modified_time: float = 0
    __should_initialized: bool = False
    __should_not_be_converted__ = True

    def __init__(
        self,
        filepath_or_data: Optional[Union[str, dict, Path]] = None,
        /,
        mode: Optional[Literal["r", "w", "a", "w=", "a="]] = None,
        *,
        filepath: Optional[Union[str, Path]] = None,
        save_on_edit: bool = False,
        read_only: Optional[Union[bool, Set[str]]] = None,
        overwrite: Optional[bool] = None,
        data: Optional[dict] = None,
        open_on_init: Optional[bool] = None,
        **kwds,
    ):
        """DH5.

        Args:
            filepath_or_data (str|dict, optional): either filepath, either data as dict.
            filepath (str|Path, optional): filepath to load. Defaults to None.
            save_on_edit (bool, optional): Save data as soon as you changed it.
                Defaults to False. And data should be saved using `save()` method.
            read_only (bool, optional): opens file in read_only mode, i.e. it cannot be modified.
                Defaults to (save_on_edit is False && overwrite is False) and filepath is set.
            overwrite (Optional[bool], optional):
                If file exists, it should be explicitly precised.
                By default raises an error if file exist.
            data (Optional[dict], optional):
                Data to load. If data provided, file . Defaults to None.
            open_on_init (Optional[bool], optional): open_on_init. Defaults to True.

        """
        if mode is not None:
            if mode.startswith("w"):
                read_only = False
            elif mode.startswith("a"):
                read_only = False
                overwrite = False
            elif mode == "r":
                read_only = True
            if mode.endswith("="):
                save_on_edit = True

        if filepath_or_data is not None and hasattr(filepath_or_data, "keys"):
            if not isinstance(filepath_or_data, dict):
                filepath_or_data = {key: filepath_or_data[key] for key in filepath_or_data.keys()}  # type: ignore
            data = data or filepath_or_data

        if isinstance(filepath_or_data, (str, Path)):
            filepath = filepath or filepath_or_data

        if filepath and not isinstance(filepath, str):
            filepath = str(filepath)

        self._data: Dict[str, Any] = data or {}
        # transform_to_possible_formats(self._data)
        self._keys: Set[str] = set(self._data.keys())
        self._last_update = set()
        self._save_on_edit = save_on_edit
        self._classes_should_be_saved_internally = set()
        self._key_prefix: Optional[str] = kwds.get("key_prefix")

        if read_only is None:
            read_only = (save_on_edit is False and not overwrite) and filepath is not None

        if open_on_init is False and overwrite is True:
            raise ValueError("Cannot overwrite file and open_on_init=False mode")
        self._open_on_init = (
            open_on_init if open_on_init is not None else (None if self._data else True)
        )
        self._unopened_keys = set()

        # if keep_up_to_data and read_only is True:
        # raise ValueError("Cannot open file in read-only and keep_up_to_data=True mode")
        # self._keep_up_to_data = keep_up_to_data

        self._read_only = read_only
        if filepath is not None:
            filepath = filepath if filepath.endswith(".h5") else filepath + ".h5"

            if (overwrite or save_on_edit) and read_only:
                raise ValueError("""Cannot open file in read_only mode and overwrite it.""")

            if os.path.exists(filepath):
                if overwrite is None and not read_only:
                    raise FileExistsError(
                        "File with the same name already exists. So you should explicitly "
                        "provide what to do with it. Set `overwrite=True` to replace file. "
                        "Set `overwrite=False` if you want to open existing file and work with it."
                    )

                if overwrite and not read_only:
                    self.__should_initialized = True
                    # os.remove(filepath)

                if read_only or (not read_only and not overwrite):
                    if self._open_on_init:
                        self._load_from_h5(filepath)
                    elif self._open_on_init is False:
                        self._keys = h5py_utils.keys_h5(filepath, key_prefix=self._key_prefix)
                        self._unopened_keys.update(self._keys)

            elif read_only:
                raise ValueError(
                    f"Cannot open file in read_only mode if file {filepath} does not exist"
                )

            # if not read_only:
            self._filepath = filepath

    @classmethod
    def open_overwrite(
        cls,
        filepath_or_data: Optional[Union[str, dict, Path]] = None,
        /,
        mode: Optional[Literal["="]] = None,
        *,
        filepath: Optional[Union[str, Path]] = None,
        save_on_edit: bool = False,
        read_only: Optional[Union[bool, Set[str]]] = None,
        overwrite: Optional[bool] = True,
        data: Optional[dict] = None,
        open_on_init: Optional[bool] = None,
        **kwds,
    ):
        """Open file in the overwrite mode.

        It deletes the file if it exists and then opens it in the write mode.
        Same syntax as `__init__` method.
        """
        mode_ = "w=" if mode == "=" else "w"
        return cls(
            filepath_or_data,
            mode=mode_,
            filepath=filepath,
            save_on_edit=save_on_edit,
            read_only=read_only,
            overwrite=overwrite,
            data=data,
            open_on_init=open_on_init,
            **kwds,
        )

    def __init__filepath__(self, *, filepath: str, filekey: str, save_on_edit: bool = False, **_):
        """Initialize a filepath. It allows to save sub DH5 objects independently.

        Args:
            filepath (str): The path to the file to be synced.
            filekey (str): The key prefix to use for the synced data.
            save_on_edit (bool, optional): Whether to save the file automatically when it is edited. Defaults to False.
            **kwargs: Additional keyword arguments to pass to the constructor.
        """
        self._filepath = filepath
        self._key_prefix = filekey
        self._save_on_edit = save_on_edit

    def _load_from_h5(
        self, filepath: Optional[str] = None, key: Optional[Union[str, Set[str]]] = None
    ) -> Set[str]:
        """Load data from h5 to self._data."""
        filepath = filepath or self._filepath
        if filepath is None:
            raise ValueError("Filepath is not specified. So cannot load_h5")
        filepath = filepath if filepath.endswith(".h5") else filepath + ".h5"
        data = h5py_utils.open_h5(filepath, key=key, key_prefix=self._key_prefix)
        self._file_modified_time = os.path.getmtime(filepath)
        return self._update(data)

    def load(self, filepath: Optional[str] = None, key: Optional[Union[str, Set[str]]] = None):
        """Load data from h5 into current object."""

        updated_from_other_file = not (filepath is None)
        updated_key = self._load_from_h5(filepath=filepath, key=key)
        if updated_from_other_file:
            for key in updated_key:
                self._keys.add(key)

        return self

    def lock_data(self: _SELF, keys: Optional[Iterable[str]] = None) -> _SELF:
        """Locks the specified keys in the database so they cannot be changed.

        Args:
            keys: An optional iterable of strings representing the keys to be locked.
            If None, all keys will be locked.

        Returns:
            A reference to the DH5 object.

        Raises:
            ValueError: If everything is already locked by read_only mode.

        Examples:
            >>> sd = DH5({"key1": "value1", "key2": "value2"})
            >>> sd.lock_data(['key1', 'key2'])
            >>> sd['key1'] = 2
            ReadOnlyKeyError: "Cannot change a read-only key 'key1'."
            >>> sd['key2'] = 5

        """
        if self._read_only is True:
            raise ValueError("Cannot lock specific data and everything is locked by read_only mode")
        if not isinstance(self._read_only, set):
            self._read_only = set()
        if keys is None:
            keys = self.keys()
        elif isinstance(keys, str):
            keys = (keys,)

        for key in keys:
            self._read_only.add(key)

        self._clean_precalculated_results()
        return self

    def unlock_data(self: _SELF, remove_keys: Optional[Iterable[str]] = None) -> _SELF:
        """Unlock the specified keys in the database so they can be changed.

        If file was opened in read-only mode you cannot unlock it, however you can open
        it again in 'a' mode and lock all keys except necessary.

        Args:
            keys: An optional iterable of strings representing the keys to be unlocked.
            If None, all keys will be unlocked.

        Returns:
            A reference to the DH5 object.

        Raises:
            ValueError: If everything is already locked by read_only mode.

        Examples:
            >>> sd = DH5({"key1": "value1", "key2": "value2"})
            >>> sd.lock_data()
            >>> sd.unlock_data('key2')
            >>> sd['key2'] = 5
            >>> sd['key1'] = 2
            ReadOnlyKeyError: "Cannot change a read-only key 'key1'."

        """
        if self._read_only is True:
            raise ValueError("Cannot unlock is global read_only mode was set to True")
        if isinstance(self._read_only, set):
            if remove_keys is None:
                self._read_only = False
            else:
                if isinstance(remove_keys, str):
                    remove_keys = (remove_keys,)
                for key in remove_keys:
                    if key in self._read_only:
                        self._read_only.remove(key)

        self._clean_precalculated_results()
        return self

    def _clean_precalculated_results(self):
        self._repr = None

    def __add_key(self, key):
        self._pre_save()
        self._keys.add(key)
        self._last_update.add(key)

    def __del_key(self, key):
        self._keys.remove(key)
        self._last_update.add(key)

    def __check_read_only_true(self, key):
        """Return true if data with this key is only available for read.

        Takes into account the external constrains.
        """
        return (self._read_only) and (self._read_only is True or key in self._read_only)

    @editing
    def update(self: _SELF, __m: Optional[dict] = None, **kwds: "DICT_OR_LIST_LIKE") -> _SELF:
        """Update data from a dictionary or keyword arguments.

        See `DH5.data_transformation` to learn more
            about how the types are converted.

        Args:
            __m (dict | None): A dictionary of key-value pairs to update the DH5 object with.
            **kwds (DICT_OR_LIST_LIKE): Keyword arguments of key-value pairs to update the DH5 object with.

        Returns:
            Self.

        Examples:
            >>> data = DH5()
            >>> data.update({'a': 1, 'b': 2})
            DH5({'a': 1, 'b': 2})
            >>> data.update(c=3, d=4)
            DH5({'a': 1, 'b': 2, 'c': 3, 'd': 4})

        """
        if __m is not None:
            kwds.update(__m)

        for key in kwds:  # pylint: disable=C0206
            if self.__check_read_only_true(key):
                raise ReadOnlyKeyError(key)
            self.__add_key(key)
            kwds[key] = transform_to_possible_formats(kwds[key])

        # self.pull(auto=True)
        self._data.update(**kwds)
        return self

    def _update(self, __m: Optional[dict] = None, **kwds: Any):
        """Update only internal data and attributes.

        Can be modified in read_only mode. Did not change a file.
        """
        if __m is not None:
            kwds.update(__m)

        for key in kwds:
            self._keys.add(key)
            self._unopened_keys.discard(key)

        self._data.update(kwds)

        return set(kwds.keys())

    @editing
    def pop(self, key: str) -> Union[Any, NotLoaded]:
        """Remove the specified key and return the value.

        Args:
            key (str): The key to remove.

        Returns:
            Same as `get` method, i.e. `DH5` if the value is a dict otherwise the value.
                If data was never loaded it will return `NotLoaded`. To load data use `get` method.

        Examples:
            >>> data = DH5({'a': 1, 'b': 2, 'c': 3})
            >>> data.pop('b')
            2

        """
        if self.__check_read_only_true(key):
            raise ReadOnlyKeyError(key, action="pop")
        self.__del_key(key)
        # self.pull(auto=True)
        if key not in self._unopened_keys:
            return self._data.pop(key)
        return NotLoaded()

    @editing
    def remove(self: _SELF, key: str) -> _SELF:
        """Remove the specified key and self.

        Args:
            key (str): The key to remove.

        Returns:
            Self.

        Examples:
            >>> data = DH5({'a': 1, 'b': 2, 'c': 3})
            >>> data.pop('b')
            DH5({'a': 1, 'c': 3})

        """
        self.pop(key)
        return self

    @overload
    def get_raw(self, __key: str) -> Any:
        """Return element as a dict. Return None if not found."""

    @overload
    def get_raw(self, __key: str, __default: _T) -> Union[Any, _T]:
        """With default value provided."""

    def get_raw(self, key: str, default: Any = None) -> Any:
        """Return raw value associated with the given key.

        Dictionaries are not converted to the `DH5` unlike `get` method.

        Args:
            key (str): Key to be searched.
            default (Any, optional): Value to be returned if the `key` is not found.
                The default value is `None`.


        Returns:
            Raw value without any conversion or the default value if the key is not found.

        Examples:
            >>> sync_data = DH5({'key1':{'a': 1, 'b': 2}, 'key2': 5})
            >>> sync_data.get_raw('key1')
            {'a': 1, 'b': 2}
            >>> sync_data.get_raw('key2')
            5
        """
        return self.__get_data__(key, default)

    @overload
    def get(self, __key: str) -> Any:
        """Return element as a DH5 class if it's dict. Return None if not found."""

    @overload
    def get(self, __key: str, __default: _T) -> Union[Any, _T]:
        """With default value provided."""

    def get(self, key: str, default: Any = None) -> Any:
        """Retrieve the value associated with the given key from the DH5 object.

        Args:
            key (str): Key to be searched.
            default (Any, optional): Value to be returned if the `key` is not found.
                The default value is `None`.

        Returns:
            The value associated with the key. If the value is a dict then it's converter
            into `DH5` object. This conversion take some time, but allow to change, save and
            this sub-object. For faster performance use `get_raw` method.

        Examples:
            >>> sync_data = DH5(filepath='data.json', data={'name': 'John', 'age': 30})
            >>> sync_data.get('name')
            'John'
            >>> sync_data.get('surname')
            None
            >>> sync_data.get('gender', 'unknown')
            'unknown' # Returns 'unknown' since 'gender' key doesn't exist

        """
        data = self.__get_data__(key, default)
        if isinstance(data, dict) and data:
            return DH5(
                filepath=self._filepath,
                data=data,
                overwrite=False,
                key_prefix=key,
                read_only=self.__check_read_only_true(key),
            )
        return data

    def __getitem__(self, __key: Union[str, tuple]) -> Any:
        """Return raw value associated with the given key.

        Same as `get_raw` but raises error if the key is not found.

        Args:
            key (str | tuple): The key to retrieve the dictionary for.

        Returns:
            Raw value without any conversion.

        Raises:
            KeyError: If the key is not found.

        """
        if isinstance(__key, tuple):
            if len(__key) > 1:
                return self.__getitem__(__key[0]).__getitem__(
                    __key[1:] if len(__key) > 2 else __key[1]
                )
            if len(__key) == 1:
                return self.__getitem__(__key[0])
            raise ValueError("Key should be a string or tuple with at least one element")
        return self.__get_data_or_raise__(__key)

    @editing
    def __setitem__(self, __key: Union[str, tuple], __value: "DICT_OR_LIST_LIKE") -> None:
        """Set value corresponding to the given key.

        See [`DH5.data_transformation`](data_transformation.md) to learn more
            about how the types are converted.

        Args:
            key (str | tuple): The key to retrieve the dictionary for.

        """
        if isinstance(__key, tuple):
            if not __key or len(__key) == 1 or len(__key) == 0:
                raise ValueError("Key should be a string or tuple with at least two elements")

            self.__add_key(__key[0])
            if self.__check_read_only_true(__key[0]):
                raise ReadOnlyKeyError(__key[0], action="set")
            return self.__getitem__(__key[0]).__setitem__(
                __key[1:] if len(__key) > 2 else __key[1], __value  # type: ignore
            )

        if self.__check_read_only_true(__key):
            raise ReadOnlyKeyError(__key, action="set")

        self.__add_key(__key)
        __value = transform_to_possible_formats(__value)

        if self._read_only is not True:
            if hasattr(__value, "save"):
                self._classes_should_be_saved_internally.add(__key)

            if hasattr(__value, "__init__filepath__") and self._filepath:
                key = __key if self._key_prefix is None else f"{self._key_prefix}/{__key}"
                __value.__init__filepath__(  # type: ignore
                    filepath=self._filepath, filekey=key, save_on_edit=self._save_on_edit
                )

            if hasattr(__value, "__post__init__"):
                __value.__post__init__()  # type: ignore

        self.__set_data__(__key, __value)
        return None

    def __delitem__(self, key: str):
        self.pop(key)

    def __getattr__(self, __name: str):
        """Call if __getattribute__ does not work."""
        if len(__name) > 1 and __name[0] == "i" and __name[1:].isdigit() and __name not in self:
            __name = __name[1:]
        if __name in self:
            data = self.get(__name)
            # if isinstance(data, dict) and data:
            #     return DH5(filepath=self._filepath, data=data, key_prefix=__name)
            return data
        raise AttributeError(f"No attribute {__name} found in DH5")

    def __setattr__(self, __name: str, __value: "DICT_OR_LIST_LIKE") -> None:
        """Call every time you set an attribute."""
        if __name.startswith("_"):
            return object.__setattr__(self, __name, __value)

        if isinstance(vars(self.__class__).get(__name), property):
            return object.__setattr__(self, __name, __value)
        return self.__setitem__(__name, __value)

    def __delattr__(self, __name: str) -> None:
        if __name in self:
            return self.__delitem__(__name)
        return object.__delattr__(self, __name)

    @overload
    def __get_data__(self, __key: str) -> Optional[None]:
        """Return None if the data doesn't contain key."""

    @overload
    def __get_data__(self, __key: str, __default: _T) -> Union[Any, _T]:
        """Return default value if the data doesn't contain key."""

    def __get_data__(self, __key: str, __default: Any = None):
        if __key in self._unopened_keys:
            self._load_from_h5(key=__key)
        data = self._data.get(__key, __default)
        if isinstance(data, NotLoaded):
            self._load_from_h5(key=__key)
            data = self._data.get(__key, __default)
        if self.__check_read_only_true(__key):
            if hasattr(data, "_read_only"):
                data._read_only = True  # type: ignore # pylint: disable=protected-access
            else:
                data = deepcopy(data)
        return data

    def __get_data_or_raise__(self, __key):
        # self.pull(auto=True)
        if __key in self._unopened_keys:
            self._load_from_h5(key=__key)
        data = self._data.__getitem__(__key)
        if isinstance(data, NotLoaded):
            self._load_from_h5(key=__key)
            data = self._data.__getitem__(__key)
        if self.__check_read_only_true(__key):
            if hasattr(data, "_read_only"):
                data._read_only = True  # type: ignore # pylint: disable=protected-access
            else:
                data = deepcopy(data)
        return data

    def __set_data__(self, __key: str, __value):
        # self.pull(auto=True)
        return self._data.__setitem__(__key, __value)

    def items(self):
        """Return all items in the collection.

        It opens all items that were not opened yet and return dictionary iterator.
        """
        if self._unopened_keys:
            self._load_from_h5(key=self._unopened_keys)
        return self._data.items()

    def values(self):
        """Return all values in the collection.

        It opens all items that were not opened yet and return dictionary iterator.
        """
        # self.pull(auto=True)
        if self._unopened_keys:
            self._load_from_h5(key=self._unopened_keys)
        return self._data.values()

    def keys(self) -> Set[str]:
        """Return all keys in the collection."""
        # self.pull(auto=True)
        return self._keys.copy().union(self._unopened_keys.copy())

    def keys_tree(self) -> Dict[str, Optional[dict]]:
        """Return dict of the keys, where value always is a dict or None.

        Examples:
            ```
            >>> sd = DH5({'a': {'b': 'value'}, 'c'})
            >>> sd.keys_tree()
            {'a': {'b': None}, 'c': None}
            ```

        For all unopened keys, it does not open them and does not explore the structure.
        """
        structure = get_keys_structure(self._data)

        for key in self._unopened_keys:
            structure[key] = None
        return structure

    def __iter__(self):
        return iter(self.keys())

    def _get_repr(self):
        if self._repr is None:
            additional_info = (
                {key: " (r)" for key in self._read_only}
                if isinstance(self._read_only, set)
                else None
            )
            self._repr = output_dict_structure(self._data, additional_info=additional_info) + (
                f"\nUnloaded keys: {self._unopened_keys}" if self._unopened_keys else ""
            )

    def __repr__(self):
        self._get_repr()

        not_saved = "" if self._last_data_saved or self._read_only is True else " (not saved)"
        mode = "r" if self._read_only is True else "w" if self._read_only is False else "rw"
        mode = "l" if self._filepath is None and self._read_only is not True else mode
        not_saved = "" if mode == "l" else not_saved

        return f"{type(self).__name__} ({mode}){not_saved}: \n {self._repr}"

    def __str__(self):
        return self.__repr__()

    def __contains__(self, item):
        return (item in self._data) or (item in self._unopened_keys)

    def __dir__(self) -> Iterable[str]:
        return list(self._keys) + self._default_attr

    def __similar__(self, other: "DH5") -> bool:
        """Check if 2 DH5 are similar.

        It means: same file, mode, save_on_edit. Does not check the data.
        """
        return (
            self._filepath == other._filepath  # pylint: disable=protected-access
            and self._read_only == other._read_only  # pylint: disable=protected-access
            and self._save_on_edit == other._save_on_edit  # pylint: disable=protected-access
            and self.__should_initialized
            == other.__should_initialized  # pylint: disable=protected-access
        )

    def _pre_save(self, *args, **kwargs):
        del args, kwargs
        if self.__should_initialized and self._filepath:
            os.remove(self._filepath)
            self.__should_initialized = False

    def save(
        self,
        only_update: Union[bool, Iterable[str]] = True,
        filepath: Optional[str] = None,
        force: Optional[bool] = None,
    ):
        """
        Save the data to a file.

        Args:
            only_update (Union[bool, Iterable[str]], optional): Determines whether to save only
                the updated data or all data.
                If True, only the updated data will be saved. If False, all data will be saved.
                If an iterable of strings is provided, only the specified keys will be saved. Defaults to True.
            filepath (str, optional): The path to the file where the data will be saved.
                If not provided, the default filepath will be used. Defaults to None.
            force (bool, optional): Determines whether to force the save operation, even if only_update is True.
                If True, the save operation will be forced. If False or None, the save operation
                will be performed according to the value of only_update. Defaults to None.

        Returns:
            self

        Raises:
            ValueError: If the file is opened in read-only mode, it cannot be saved.
                The file should be reopened in write mode before saving.
        """
        if self._read_only is True:
            raise ValueError("Cannot save opened in a read-only mode. Should reopen the file")

        self._pre_save()

        if force is True or filepath is not None:
            only_update = False

        if isinstance(only_update, Iterable):
            last_update = self._last_update.intersection(only_update)
            self._last_update = self._last_update.difference(only_update)
        else:
            last_update, self._last_update = self._last_update, set()

        if len(self._last_update) == 0:
            self._last_data_saved = True

        filepath = self._check_if_filepath_was_set(filepath, self._filepath)

        if only_update is False:
            if self._read_only is False:
                data_to_save = self._data
            else:
                data_to_save = {
                    key: value
                    for key, value in self._data.items()
                    if key not in self._read_only or key in last_update
                }
            data_to_save.update(
                {
                    key: None
                    for key in last_update
                    if (key not in self._data) and not self.__check_read_only_true(key)
                }
            )

            self.__h5py_utils_save_dict_with_retry(filepath=filepath, data=data_to_save)

            return self

        for key in last_update:
            if key in self._classes_should_be_saved_internally:
                obj = self._data[key]
                if hasattr(obj, "save"):
                    self._data[key].save(only_update=only_update)
                else:
                    self._classes_should_be_saved_internally.remove(key)

        self.__h5py_utils_save_dict_with_retry(
            filepath=filepath,
            data={
                key: self._data.get(key)
                for key in last_update
                if key not in self._classes_should_be_saved_internally
            },
        )

        return self

    def __h5py_utils_save_dict_with_retry(self, filepath: str, data: dict):
        # print("open", time.time(), self._raise_file_locked_error)
        for i in range(self._retry_on_file_locked_error):
            try:
                # print("_raise_file_locked_error", self._raise_file_locked_error, list(data.keys()))
                self._file_modified_time = h5py_utils.save_dict(
                    filename=filepath + ".h5", data=data, key_prefix=self._key_prefix
                )
                return
            except h5py_utils.FileLockedError as error:
                if self._raise_file_locked_error:
                    raise error
                logging.info("File is locked. waiting 2s and %d more retrying.", i)
                from ..utils import async_utils

                async_utils.sleep(1)

        raise h5py_utils.FileLockedError(
            f"Even after {self._retry_on_file_locked_error} data was not saved"
        )

    @staticmethod
    def _check_if_filepath_was_set(filepath: Optional[str], filepath2: Optional[str]) -> str:
        """Return path to the file with filename, but without extension."""
        filepath = filepath or filepath2
        if filepath is None:
            raise ValueError("Should provide filepath or set self.filepath before saving")
        filepath = (filepath.rsplit(".h5", 1)[0]) if filepath.endswith(".h5") else filepath
        return filepath

    @property
    def filepath(self):
        """
        Return the filepath without the '.h5' extension.

        If the filepath is None, returns None.
        """
        return None if self._filepath is None else (self._filepath.rsplit(".h5", 1)[0])

    @filepath.setter
    def filepath(self, value: str):
        if not isinstance(value, str):
            value = str(value)
        self._filepath = value if value.endswith(".h5") else value + ".h5"

    @property
    def filename(self) -> Optional[str]:
        """
        Return the filename of the current filepath without '.h5' extension.

        Returns:
            Optional[str]: The filename of the current filepath, or None if the filepath is None.
        """
        filepath = self.filepath
        if filepath is None:
            return None
        return os.path.basename(filepath)

    @property
    def save_on_edit(self):
        """Return the current value of the save_on_edit attribute."""
        return self._save_on_edit

    def asdict(self):
        """
        Return the internal data of the object as a dictionary.

        Returns:
            dict: A dictionary representation of the object's internal data.
        """
        return self._data

    def pull_available(self):
        """
        Check if the file has been modified elsewhere since the last save.

        Raises:
            ValueError: If the filepath has not been set.

        Returns:
            bool: True if the file has been modified, False otherwise.
        """
        if self.filepath is None:
            raise ValueError("Cannot pull from file if it's not been set")
        file_modified = os.path.getmtime(self.filepath + ".h5")
        return self._file_modified_time != file_modified

    def pull(self, force_pull: bool = False):
        """
        Pull data from a file and reloads it into the object.

        Args:
            force_pull (bool, optional): If True, forces to update data even if the file
            has not been modified. Defaults to False.

        Raises:
            ValueError: If the filepath has not been set.

        Returns:
            self: The updated object.
        """
        if self.filepath is None:
            raise ValueError("Cannot pull from file if it's not been set")

        if force_pull or self.pull_available():
            logging.debug("File modified so it will be reloaded.")
            self._data = {}
            self._keys = set()
            self._clean_precalculated_results()
            self._load_from_h5()

        return self

    @overload
    def close_data(self, key: None = None, every: Literal[True] = True):
        """Close every opened key so it could be collected by the garbage collector afterwards."""

    @overload
    def close_data(self, key: Iterable[str]):
        """Close every key provided so it could be collected by the garbage collector afterwards."""

    @overload
    def close_data(self, key: str):  # type: ignore
        """Close the key so it could be collected by the garbage collector afterwards."""

    def close_data(
        self,
        key: Optional[Union[str, Iterable[str]]] = None,
        every: Optional[Literal[True]] = None,
    ):
        """Close the key so it could be collected by the garbage collector afterwards.

        Args:
            key (str | Iterable[str], optional): key or keys that should be closed. Defaults to None.
            every (True, optional): put to True if all keys should be closed. Defaults to None.

        Raises:
            ValueError: if both key and every are not provided.

        Returns:
            Self.
        """
        if every is True:
            for k in self.keys():
                self.close_data(k)
            return self
        elif key is None:
            raise ValueError("Should provide key or every=True.")

        if not isinstance(key, str):
            for k in key:
                self.close_data(k)
            return self

        if key not in self._unopened_keys:
            self._data.pop(key)
        self._unopened_keys.add(key)

        return self



"""This module contains asynchronous functions."""
import threading


def sleep(delay):
    """
    Sleep for the specified delay in seconds.

    Args:
        delay (float): The delay in seconds.

    """
    event = threading.Event()
    event.wait(delay)
    event.clear()



"""Utils to save and load h5 files."""

import os
from typing import Literal, Optional, Set, Union

import h5py
import numpy as np

from ..errors import FileLockedError
from ..types import ClassWithAsdict
from .data_transformation import transform_not_dict_on_save, transform_on_open


class LockFile:
    """Context manager that creates a lock file during writing to prevent conflicts.

    Usage:
        To use the LockFile context manager, simply wrap your code that writes to a file with it.
        This will ensure that only one process can write to the file at a time.

        Example:
        ```
        with LockFile("data.txt"):
             # Code that writes to the file
             with open("data.txt", "w") as f:
                 pass
        ```

    """

    def __init__(self, filename):
        """Initialize the LockFile object.

        Args:
            filename (str): The filename or filepath. Can end with an extension.
        """
        self.lock_filename = os.path.splitext(filename)[0] + ".lock"

    def __enter__(self):
        """Enter the context and create the lock file."""
        if os.path.exists(self.lock_filename):
            raise FileLockedError("File locked and cannot be opened in write mode")
        with open(self.lock_filename, "w", encoding="utf-8"):
            pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the context and remove the lock file."""
        if os.path.exists(self.lock_filename):
            os.remove(self.lock_filename)


# -------------- Open file ----------------


def open_h5_group(
    group: Union[h5py.File, h5py.Group],
    key: Optional[Union[str, Set[str]]] = None,
    key_prefix: Optional[str] = None,
) -> dict:
    """Open h5 group and return dict.

    Args:
        group (Union[h5py.File, h5py.Group]): group to open
        key (Optional[Union[str, Set[str]]], optional): Keys to load. Defaults to None.
        key_prefix (Optional[str], optional): Key prefix. Defaults to None.

    Returns:
        dict: The loaded data. Each value was transform using transform_on_open function.
    """
    data = {}
    if key is not None:
        key = key if isinstance(key, set) else set([key])

    for group_key in group.keys():
        key = key if key_prefix is None else f"{key_prefix}/{key}"
        if key is not None and group_key not in key:
            continue
        value = group[group_key]
        if isinstance(value, h5py.Group):
            data[group_key] = open_h5_group(value)
        else:
            data[group_key] = transform_on_open(value[()])  # type: ignore
    return data


def open_h5(
    fullpath: str, key: Optional[Union[str, Set[str]]] = None, key_prefix: Optional[str] = None
) -> dict:
    """Open h5 file and return dict.

    Args:
        fullpath (str): Full filepath to the file.
        key (str | set[str], optional): Key or keys to load. Defaults to all keys.
        key_prefix (str, optional): Key prefix to put before each key. Defaults to None.

    Returns:
        dict: The loaded data. Each value was transform using transform_on_open function.
    """
    with h5py.File(fullpath, "r") as file:
        if key_prefix is None:
            return open_h5_group(file, key=key)
        return open_h5_group(file[key_prefix], key=key)  # type: ignore


# -------------- Save file ----------------


def save_sub_dict(
    group: Union[h5py.File, h5py.Group],
    data: Union[dict, list, np.ndarray, ClassWithAsdict],
    key: str,
    use_compression: Optional[Union[Literal[True], str]] = None,
):
    """Save the dict to a group. Each object is converted to a dict, array or simple value.

    Args:
        group (Union[h5py.File, h5py.Group]): h5 group where to save.
        data (Union[dict, list, np.ndarray, ClassWithAsdict]): Any object that either has 'asdict'
         or 'asarray' method, either is a dict, list or np.ndarray, either can be transformed
         using transform_not_dict_on_save function.
        key (str): Desire key for the data inside h5 file .
        use_compression (str|True, optional): If compression should be used.
         If true, 'gzip' is used, otherwise you can specify compression by providing a str.
         Defaults to not compressed.
    """
    if hasattr(data, "asdict"):
        data = data.asdict()  # type: ignore
    if hasattr(data, "asarray"):
        data = data.asarray()  # type: ignore
    if isinstance(data, dict):
        g = group.create_group(key)
        for k, v in data.items():
            save_sub_dict(g, v, k, use_compression=use_compression)
    elif (key is not None) and (data is not None):
        data = transform_not_dict_on_save(data)  # type: ignore
        if isinstance(data, (np.ndarray, list)):
            use_compression = "gzip" if use_compression is True else use_compression
            group.create_dataset(key, data=data, compression=use_compression)  # compression="gzip"
        else:
            group.create_dataset(key, data=data)


def save_dict(
    filename: str,
    data: dict,
    key_prefix: Optional[str] = None,
    use_compression: Optional[Union[Literal[True], str]] = None,
) -> float:
    """Save dict to h5 file.

    Args:
        filename (str): Full filepath to the file.
        data (dict): Data to save.
        key_prefix (str, optional): Key prefix of the desired location inside h5 file.
         If nested use `key1/key2`. Defaults to None, i.e. root of the file.
        use_compression (str|True, optional): If compression should be used. If true,
         'gzip' is used, otherwise you can specify compression by providing a str.
         Defaults to not compressed.

    Returns:
        float: Time of the last modification of the file.

    Example:
        ```
        save_dict('/path/to/file.h5', {'key1': 1})
        ```



    """
    dirname = os.path.dirname(filename)
    if dirname and not os.path.exists(dirname):
        os.makedirs(dirname, exist_ok=True)

    mode = "a"  # if os.path.exists(filename) else 'w'
    with LockFile(filename), h5py.File(filename, mode) as file:
        for key, value in data.items():
            key = key if key_prefix is None else f"{key_prefix}/{key}"
            if key in file:  # .keys():
                file.pop(key)
            if value is None:
                continue
            save_sub_dict(file, value, key, use_compression=use_compression)
    return os.path.getmtime(filename)


# -------------- Load keys ----------------


# /doc add example to this doc
def keys_h5(filename, key_prefix: Optional[str] = None) -> Set[str]:
    """
    Return the keys of an h5 file.

    Args:
        filename (str): Full filepath to the h5 file.
        key_prefix (str, optional): Key prefix to look at. If nested, use `key1/key2`.
            Defaults to None.

    Returns:
        Set[str]: Set of the keys in the h5 file.

    Example:
        ```
        >>> keys_h5('/path/to/file.h5', 'group1')
        {'key1', 'key2', 'key3'}
        ```
    """
    with h5py.File(filename, "r") as file:
        if key_prefix is not None:
            file = file[key_prefix]
        return set(file.keys())  # type: ignore


# -------------- Delete keys ----------------


def del_dict(
    filename: str,
    key: str,
    key_prefix: Optional[str] = None,
) -> float:
    """
    Delete a key from an h5 file.

    Args:
        filename (str): Full filepath to the file.
        key (str): The key to delete. If nested, use `key1/key2`.
        key_prefix (str, optional): If provided, the key prefix to use. Defaults to None.

    Returns:
        float: The time of the last modification of the file.

    Example:
        ```
        >>> del_dict('/path/to/file.h5', 'key1', 'prefix')
        "some_value"
        ```
    """
    with LockFile(filename), h5py.File(filename, "a") as file:
        key = key if key_prefix is None else f"{key_prefix}/{key}"
        file.pop(key)
    return os.path.getmtime(filename)



"""Random utilities for the dh5 package."""
from typing import Optional, Tuple, Union, TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path  # pragma: no cover


def get_timestamp() -> str:
    """Get the current timestamp in the format 'YYYY_MM_DD__HH_MM_SS'.

    Returns:
        str: The current timestamp.
    """
    import datetime

    x = datetime.datetime.now()
    return x.strftime("%Y_%m_%d__%H_%M_%S")


def lstrip_int(line: str) -> Optional[Tuple[str, str, str]]:
    """Find whether timestamp ends.
    Returns timestamp and rest of the line, if possible.

    Args:
        line (str): The line to check for a timestamp.

    Returns:
        Optional[Tuple[str, str, str]]: A tuple containing the prefix, main part, and suffix
            of the line if a timestamp is found, None otherwise.
    """
    import re

    main = re.search("_[A-Za-z]", line)
    if main is None:
        return None
    prefix, main = line[: main.start()], line[main.start() + 1 :]
    if "__" in main:
        suffix_index = main.rfind("__")
        suffix = main[suffix_index + 2 :]
        if suffix.isdecimal():
            main = main[:suffix_index]
        else:
            suffix = ""
    else:
        suffix = ""

    if not prefix.replace("_", "").replace("-", "").isdigit():
        return None

    return prefix.strip("_"), main, suffix


def get_path_from_filename(filename: Union[str, "Path"]) -> Union[str, tuple]:
    """
    Given a filename, returns the path to the file.

    If the filename contains a path, returns the filename as is.
    If the filename ends with '.h5', removes the extension.
    If the filename starts with a timestamp, returns a tuple with (the suffix that does after
    the timestamp, the full filename without '.h5' extension in the end).

    Args:
        filename (Union[str, "Path"]): The filename to get the path from.

    Returns:
        Union[str, tuple]: The path to the file as tuple (folder, filename) or the filename itself.
        Return directly filename if it contains a slash, therefore represents the path.
    """
    filename = str(filename)

    if "/" in filename or "\\" in filename:
        return filename

    filename = filename[:-3] if filename.endswith(".h5") else filename

    name_with_prefix = lstrip_int(filename)

    if name_with_prefix:
        suffix = name_with_prefix[1]
        return (suffix, filename)
    return filename



# flake8: noqa: D100
import pathlib
import os
from typing import Iterable, Optional, Union


class Path(type(pathlib.Path())):
    """Make pathlib.Path better.

    Functionality added:
    - sum 2 path or path and str. Action: sum 2 str
    - makedirs. Action: os.makedirs
    - make_extension
    - dirname
    - basename
    """

    def __add__(self, other: Union["Path", str]):
        """Sum two paths as concatenation of two strings."""
        if not isinstance(other, str):
            other = str(other)
        return type(self)(self.as_posix() + other)

    def makedirs(self, mode=0o777, exist_ok=False):
        """Create a directories if needed.

        If self is a file creates directories to dirname(self)."""
        file = os.path.dirname(self) if "." in os.path.basename(self) else self
        if not os.path.exists(file):
            os.makedirs(file, mode=mode, exist_ok=exist_ok)
        return self

    def make_extension(self, extension: Union[Iterable[str], str]):
        """Make sure that the extension is the one provided.

        If the list of extensions provided, then any extensions are allowed,
        but is no one satisfied, uses first to determine extension.

        Examples:
        --------
        >>> path = Path("some_file.json")
        >>> path.make_extension(".json") # -> "some_file.json"
        >>> path.make_extension(["txt", "csv"]) # -> "some_file.json.txt"

        """
        if isinstance(extension, str):
            extension = [extension]
        if not isinstance(extension, Iterable):
            raise ValueError("Extension must be a string or an iterable")
        extension = [ext if ext.startswith(".") else ("." + ext) for ext in extension]
        path = self.as_posix()
        for ext in extension:
            if path.endswith(ext):
                return self
        return type(self)(self.as_posix() + extension[0])

    @property
    def dirname(self) -> "Path":
        """Same as os.path.dirname."""
        return type(self)(os.path.dirname(self))

    @property
    def basename(self) -> "Path":
        """Same as os.path.basename."""
        return type(self)(os.path.basename(self))

    @property
    def str(self) -> str:
        return str(self)


def get_file_path(
    filename: Union[str, Path],
    *,
    path: Optional[Union[str, Path]] = None,
    extension: Optional[str] = None
) -> Path:
    """Get a Path for a file called `filename` at `path` location with `extension`."""
    filename = Path(filename)
    if extension:
        filename = filename.make_extension(extension)
    if path is not None:
        return Path(path, filename)
    return Path(filename)



# flake8: noqa: D101, D102, D107
"""Module contains the errors used inside DH5."""


from typing import Optional, Set, Union


class FileLockedError(Exception):
    """Exception raised when a file is locked."""


class ReadOnlyKeyError(KeyError):
    """Exception raised when a key or set of keys is read-only."""

    def __init__(self, key: Optional[Union[str, Set[str]]] = None, action="change"):
        if isinstance(key, str):
            self.message = f"Cannot {action} a read-only key '{key}'."
        elif isinstance(key, (list, tuple, set)):  # pragma: no cover
            self.message = f"Cannot {action} a read-only keys {key}."  # pragma: no cover
        else:  # pragma: no cover
            self.message = f"Cannot {action} a read-only object."  # pragma: no cover
        super().__init__(self.message)



"""Internal classes for DH5."""


class NotLoaded:
    """Data that has not been loaded yet."""

    def __init__(self):
        """Nothing to do."""

    def __str__(self) -> str:
        return """This key is not loaded. If you see this message,
                it means that you accessed the key not via DH5 object"""  # pragma: no cover

    def __repr__(self) -> str:
        return """This key is not loaded"""  # pragma: no cover



"""Output the structure of a dictionary in a pretty way."""


import json
from typing import Dict, Optional

import numpy as np


def output_dict_structure(data: dict, additional_info: Optional[Dict[str, str]] = None) -> str:
    """
    Convert a dictionary into a JSON-like string representation of its structure.

    Args:
        data (dict): The input dictionary.
        additional_info (Optional[Dict[str, str]]): Additional information to be added to the string representation.
            Each key-value pair in the additional_info dictionary will be appended to the corresponding key in the
            string representation. The key will be enclosed in double quotes and the value will be appended without
            quotes.

    Returns:
        str: The JSON-like string representation of the dictionary structure.
    """
    dict_str = dict_to_json_format_str(get_dict_structure(data))
    if additional_info:
        for key, value in additional_info.items():
            dict_str = dict_str.replace(f'"{key}":', f'"{key}"{value}:')
    return dict_str


def dict_to_json_format_str(data: dict) -> str:
    """
    Convert a dictionary to a JSON formatted string.

    Args:
        data (dict): The dictionary to be converted.

    Returns:
        str: The JSON formatted string representation of the dictionary.
    """
    return json.dumps(data, sort_keys=True, indent=4)


def get_dict_structure(data: dict, level: int = 3) -> dict:
    """
    Recursively analyzes the structure of a dictionary.
    Returns a dictionary containing information about the types and shapes of the values.

    Args:
        data (dict): The dictionary to analyze.
        level (int, optional): The maximum depth to analyze the dictionary. Defaults to 3.

    Returns:
        dict: A dictionary containing information about the structure of the input dictionary.
    """
    structure = {}

    for k, v in data.items():
        if isinstance(v, dict):
            if level:
                internal_structure = (
                    get_dict_structure(v, level=level - 1) if len(v) else "empty dict"
                )
                structure[k] = (
                    "variable of type dict" if len(internal_structure) > 5 else internal_structure
                )
            else:
                structure[k] = "variable of type dict"

        elif isinstance(v, (np.ndarray, list)):
            structure[k] = f"shape: {np.shape(v)} (type: {type(v).__name__})"
        elif isinstance(v, (int, np.int_)):  # type: ignore
            structure[k] = f"{v:.0f} (type : {type(v).__name__})"
        elif isinstance(v, (float, np.float_, complex, np.complex_)):  # type: ignore
            str_value = f"{v:.3f}" if 0.1 <= abs(v) <= 100.0 else f"{v:.3e}"
            structure[k] = f"{str_value} (type : {type(v).__name__})"
        else:
            structure[k] = f"variable of type {type(v).__name__}"
    return structure


def get_keys_structure(data) -> dict:
    """
    Recursively traverses a dictionary and returns its structure with keys and None values.

    Args:
        data (dict): The dictionary to analyze.

    Returns:
        dict: A dictionary representing the structure of the input dictionary, with keys and None values.

    """
    structure = {}
    for k, v in data.items():
        if isinstance(v, dict):
            structure[k] = get_keys_structure(v)
        else:
            structure[k] = None

    return structure
